/*
 * Copyright 1992 David Lemke and Network Computing Devices
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Network Computing Devices not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Network Computing
 * Devices makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * NETWORK COMPUTING DEVICES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL NETWORK COMPUTING DEVICES BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors:
 *		Dave Lemke
 *		lemke@ncd.com
 *
 *		Osborne Hardison
 *		hardison@ncd.com
 *
 *		Network Computing Devices, Inc
 *		350 North Bernardo Ave
 *		Mountain View, CA 94043
 *
 *	%W%	%E%
 *
 */

/*
 * zap -- visually mangle copy of screen, window by window
 */

#ifdef hpux
#define srandom srand
#define random rand
#endif


#ifndef vms
#define rnd(x)  (random() % (x))
#else
#define rnd(x)	(rand() & x)
#endif				/* vms */




#define min(a, b)       ((a) < (b) ? (a) : (b))
#define max(a, b)       ((a) > (b) ? (a) : (b))
#define abs(a)          ((a) < 0 ? -(a) : (a))

#include	<stdio.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/keysym.h>
#include	"vroot.h"
#include	"zilla_f.xbm"
#include	"zilla_l1.xbm"
#include	"zilla_r1.xbm"
#include	"zilla_l2.xbm"
#include	"zilla_r2.xbm"
#include	"zilla_l3.xbm"
#include	"zilla_r3.xbm"
#include	"zilla_u1.xbm"
#include	"zilla_u2.xbm"

#define	NUM_ZILLAS	9
#define	ZILLA_WIDTH	300
#define	ZILLA_HEIGHT	300


static Display *dpy;
static Window win;
static GC   zap_gc,
            zilla_gc;
static unsigned long black,
            white;
static int  screen;
static unsigned int  num_windows;
static Region full_region;
static Region *region_list;

static Pixmap zilla_map[NUM_ZILLAS];

static Bool debug = False;

static char *progname;

static void find_all_windows();
static void zap_all_windows();
static void zap_delay();

/* various types of window zaps */
#define	ZAP_RANDOM	-1
#define	ZAP_CLEAR_WIN	0
#define	ZAP_FLING	1
#define	ZAP_TV_OFF	2
#define	ZAP_WIPE	3
#define	ZAP_ZILLA	4

/* TBD */
/* sides fall in */
#define	ZAP_COLLAPSE	5
/* 4 sided wipe */
#define	ZAP_DWINDLE	6


static struct {
    int         type;
    char       *name;
}           modes[] = {

    {
	ZAP_CLEAR_WIN, "clear"
    },
    {
	ZAP_FLING, "fling"
    },
    {
	ZAP_TV_OFF, "tv"
    },
    {
	ZAP_WIPE, "wipe"
    },
    {
	ZAP_ZILLA, "zilla"
    },
    {
	ZAP_RANDOM, "random"
    },
};

#define	NUM_ZAP_TYPES	(sizeof(modes)/sizeof(modes[0]))

static int
strtomode(str)
    char       *str;
{
    int         i;

    for (i = 0; i < NUM_ZAP_TYPES; i++) {
	if (!strcmp(str, modes[i].name))
	    return modes[i].type;
    }
    return ZAP_RANDOM;
}

static void
usage()
{
    int         i;

    fprintf(stderr, "Usage: %s [-display displayname] [-dont_grab] [-mode ",
	    progname);
    for (i = 0; i < NUM_ZAP_TYPES; i++) {
	fprintf(stderr, "%s ", modes[i].name);
    }
    fprintf(stderr, "]\n");

    exit(1);
}

main(argc, argv)
    int         argc;
    char      **argv;
{
    char       *displayname = NULL;
    XSetWindowAttributes xswat;
    unsigned long wmask,
                gcmask;
    int         i;
    Bool        dont_grab = False;
    XGCValues   gcvals;
    char       *modetype = NULL;
    int         mode = ZAP_ZILLA;

    progname = argv[0];
    for (i = 1; i < argc; i++) {
	if (!strncmp(argv[i], "-di", 3)) {
	    if (argv[i + 1])
		displayname = argv[++i];
	    else
		usage();
	} else if (!strncmp(argv[i], "-de", 3)) {
	    debug = True;
	} else if (!strncmp(argv[i], "-do", 3)) {
	    dont_grab = True;
	} else if (!strncmp(argv[i], "-mo", 3)) {
	    if (argv[i + 1])
		modetype = argv[++i];
	    else
		usage();
	} else
	    usage();
    }

    dpy = XOpenDisplay(displayname);
    if (!dpy) {
	fprintf(stderr, "Can't open display %s\n",
		(displayname ? displayname : ""));
	exit(-1);
    }
    screen = DefaultScreen(dpy);

#ifndef SABER

#ifndef vms
    srandom(getpid());
#else
    srand(getpid());
#endif				/* vms */

#else
    srandom((long) time((time_t) NULL));
#endif

    XSynchronize(dpy, debug);

    if (!dont_grab)
	XGrabServer(dpy);

    find_all_windows();

    xswat.override_redirect = True;
    xswat.do_not_propagate_mask = KeyReleaseMask |
	ButtonPressMask | ButtonReleaseMask;
    xswat.event_mask = KeyPressMask;
    wmask = CWOverrideRedirect | CWDontPropagate | CWEventMask;
    win = XCreateWindow(dpy, RootWindow(dpy, screen), 0, 0,
			DisplayWidth(dpy, screen), DisplayHeight(dpy, screen),
			0, CopyFromParent, CopyFromParent, CopyFromParent,
			wmask, &xswat);

    black = BlackPixel(dpy, screen);
    white = WhitePixel(dpy, screen);

    gcvals.graphics_exposures = False;
    gcvals.fill_style = FillOpaqueStippled;
    gcvals.foreground = white;
    gcvals.background = black;
    gcmask = GCGraphicsExposures | GCFillStyle |
	GCForeground | GCBackground;
    zap_gc = XCreateGC(dpy, win, gcmask, &gcvals);

    if (modetype)
	mode = strtomode(modetype);

    if (mode == ZAP_ZILLA || mode == ZAP_RANDOM) {
	zilla_map[0] = XCreatePixmapFromBitmapData(dpy, win, zilla_f_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
	zilla_map[1] = XCreatePixmapFromBitmapData(dpy, win, zilla_l1_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
	zilla_map[2] = XCreatePixmapFromBitmapData(dpy, win, zilla_l2_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
	zilla_map[3] = XCreatePixmapFromBitmapData(dpy, win, zilla_l3_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
	zilla_map[4] = XCreatePixmapFromBitmapData(dpy, win, zilla_r1_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
	zilla_map[5] = XCreatePixmapFromBitmapData(dpy, win, zilla_r2_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
	zilla_map[6] = XCreatePixmapFromBitmapData(dpy, win, zilla_r3_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
	zilla_map[7] = XCreatePixmapFromBitmapData(dpy, win, zilla_u1_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
	zilla_map[8] = XCreatePixmapFromBitmapData(dpy, win, zilla_u2_bits,
					       zilla_f_width, zilla_f_height,
						   1, 0, 1);
    }
    gcmask = GCFillStyle | GCForeground | GCBackground;
    gcvals.fill_style = FillOpaqueStippled;
    gcvals.foreground = black;
    gcvals.background = white;
    zilla_gc = XCreateGC(dpy, win, gcmask, &gcvals);

    XMapWindow(dpy, win);
    zap_delay();

    zap_all_windows(mode);
    free(region_list);

    if (!dont_grab)
	XUngrabServer(dpy);
    exit(0);
}

static Bool
check_escape_hatch()
{
    XEvent      ev;

    while (XPending(dpy)) {
	XNextEvent(dpy, &ev);
	if (ev.type == KeyPress &&
		(XLookupKeysym((XKeyEvent *) & ev, 0) == XK_Delete))
	    return True;
    }
    return False;
}

static      Region
compute_full_region(list, num)
    Region     *list;
    int         num;
{
    int         i;
    Region      total;

    total = XCreateRegion();
    for (i = 0; i < num; i++) {
	if (!list[i])
	    continue;
	XUnionRegion(total, list[i], total);
    }
    return total;
}

static      Region
rect_to_region(x, y, w, h)
    int         x,
                y,
                w,
                h;
{
    XPoint      pts[4];
    Region      new;

    pts[0].x = x;
    pts[0].y = y;
    pts[1].x = x + w;
    pts[1].y = y;
    pts[2].x = x + w;
    pts[2].y = y + h;
    pts[3].x = x;
    pts[3].y = y + h;
    new = XPolygonRegion(pts, 4, WindingRule);

    return new;
}


static void
find_all_windows()
{
    Window      root,
                parent;
    Window     *kids;
    int         i,
                good;
    XWindowAttributes wattrs;

    if (!XQueryTree(dpy, RootWindow(dpy, screen), &root, &parent,
		    &kids, &num_windows)) {
	fprintf(stderr, "Couldn't get window tree - aborting\n");
	exit(-1);
    }
    region_list = (Region *) calloc(sizeof(Region) * num_windows, 1);
    if (!region_list) {
	fprintf(stderr, "Out of memory\n");
	exit(-1);
    }
    for (i = 0, good = 0; i < num_windows; i++) {
	if (!XGetWindowAttributes(dpy, kids[i], &wattrs))
	    continue;
	/* ignore unmmaped/hidden windows */
	if (wattrs.map_state != IsViewable)
	    continue;
	/* and InputOnly */
	if (wattrs.class == InputOnly)
	    continue;
	/* what *is* this? */
	if (wattrs.x + wattrs.width < 0 && wattrs.y + wattrs.height < 0)
	    continue;
	region_list[good++] = rect_to_region(wattrs.x, wattrs.y,
					     wattrs.width, wattrs.height);
    }
    XFree(kids);
    num_windows = good;
    full_region = compute_full_region(region_list, num_windows);
    if (debug)
	fprintf(stderr, "found %d windows\n", num_windows);
}

/*
 * fling area off screen
 */
static void
fling_area(x, y, w, h, dx, dy)
    int         x,
                y,
                w,
                h,
                dx,
                dy;
{
    int         adx,
                ady;

    /* make sure it goes somewhere... */
    if (dx == 0)
	dx = 1;
    if (dy == 0)
	dy = -1;
    if (debug)
	fprintf(stderr, "flinging %dx%d @ %d %d by %d %d\n",
		w, h, x, y, dx, dy);
    adx = abs(dx);
    ady = abs(dy);
    while (1) {
	/* time to kick out? */
	if (y > (DisplayHeight(dpy, screen) + h + 2) || ((y < 0) && (y < -h - 2)))
	    break;
	else if (x > (DisplayWidth(dpy, screen) + w + 2) || ((x < 0) && (x < -w - 2)))
	    break;
	XCopyArea(dpy, win, win, zap_gc, x, y, w, h,
		  x + dx, y + dy);
	if (dx < 0) {
	    XFillRectangle(dpy, win, zap_gc, x + w, y, adx, h + adx);
	} else {
	    XFillRectangle(dpy, win, zap_gc, x, y, dx, h);
	}
	if (dy < 0) {
	    XFillRectangle(dpy, win, zap_gc, x, y + h, w + ady, ady);
	} else {
	    XFillRectangle(dpy, win, zap_gc, x, y, w, dy);
	}
	x += dx;
	y += dy;
    }
}

/*
 * routine to "walk" zilla over to a specific point on the screen and
 * stomp or kick the adjoining window. OBH
 */
static void
walkZilla(zmap, x, y)
    Pixmap      zmap;
    int         x;
    int         y;
{
    Pixmap      tmap = zmap;
    static int  cx = -ZILLA_WIDTH;
    static int  cy = 200;
    int         i = 0;

    if (debug)
	fprintf(stderr, "Zilla walk: %d %d to %d %d\n", cx, cy, x, y);

    if (cx) {
	while ((cx != x) || (cy != y)) {
	    i++;
	    if (cy != y)
		cy += (cy < y ? 1 : -1);
	    if (cx != x) {
		if (cx > x) {
		    cx += -1;
		    if (i % 2)
			tmap = zilla_map[1];
		    else
			tmap = zilla_map[2];
		} else {
		    cx += 1;
		    if (i % 2)
			tmap = zilla_map[4];
		    else
			tmap = zilla_map[5];
		}
	    } else {
		if (i % 2)
		    tmap = zilla_map[7];
		else
		    tmap = zilla_map[8];
	    }
	    XCopyPlane(dpy, tmap, win, zilla_gc, 0, 0,
		       ZILLA_WIDTH, ZILLA_HEIGHT, cx, cy, 1);
	}
    } else {
	cx = x;
	cy = y;
    }
    XCopyPlane(dpy, zmap, win, zilla_gc, 0, 0, ZILLA_WIDTH, ZILLA_HEIGHT,
	       x, y, 1);
}

/*
 * draw Godzilla doing the fling
 */
static void
zilla(x, y, w, h)
    int         x,
                y,
                w,
                h;
{
    int         delta;
    int         zx,
                zy,
                dx,
                dy;
    Bool        draw_zilla = True;
    Pixmap      zmap;

    /* set reasonable defaults */
    zmap = zilla_map[0];

    /* figure out where zilla should stand and which way he should fling */
    if (x > ZILLA_WIDTH / 2) {	/* stand at left */
	zx = x - ZILLA_WIDTH;
	zy = y;
	zmap = zilla_map[6];
	dx = 10;
	dy = 3;
    } else if (DisplayWidth(dpy, screen) - (x + w) > ZILLA_WIDTH / 2) {
	zx = x + w;
	zy = y;
	zmap = zilla_map[3];
	dx = -10;
	dy = -3;
    } else if (y > ZILLA_HEIGHT) {	/* stand above */
	zy = y - ZILLA_HEIGHT;
	zx = (x + (w / 2));
	dx = 0;
	dy = 10;
    } else if ((DisplayHeight(dpy, screen) - h + y) > ZILLA_HEIGHT) {	/* below */
	zy = DisplayHeight(dpy, screen) - h + y;
	zx = (x + (w / 2));
	dx = 0;
	dy = -10;
    } else {
	dx = (rnd(20) - 10) * 2;
	dy = (rnd(20) - 10) * 2;
	draw_zilla = False;
	printf("not drawing zilla\n");
    }
    if (zy + ZILLA_HEIGHT > DisplayHeight(dpy, screen))
	zy = DisplayHeight(dpy, screen) - ZILLA_HEIGHT;

    if (draw_zilla)
	walkZilla(zmap, zx, zy);
    fling_area(x, y, w, h, dx, dy);
}

#define	TV_INC	3
#define	TV_SAVE	3

/*
 * imitiate TV being turned off -- shrink Y, then X
 */
static void
tv_off(x, y, w, h)
    int         x,
                y,
                w,
                h;
{
    int         i,
                dy,
                dx,
                newh,
                lasty;


    if (debug)
	fprintf(stderr, "tvzap %dx%d @ %d %d\n", w, h, x, y);
    dy = TV_INC;

    /* TV doesn't look right unless it has a black BG */
    XSetFillStyle(dpy, zap_gc, FillSolid);
    XSetForeground(dpy, zap_gc, black);
    i = 0;
    while (1) {
	if ((i * dy) > (h / 2 - TV_SAVE))
	    break;
	XFillRectangle(dpy, win, zap_gc, x, y + i * dy, w, dy);
	i++;
	XFillRectangle(dpy, win, zap_gc, x, y + h - (i * dy), w, dy);
	XSync(dpy, 0);
    }
    lasty = i + y;

    dx = TV_INC;
    newh = h - i * 2 + TV_SAVE;
    i = 0;
    while (1) {
	if ((i * dx) > (w / 2))
	    break;
	XFillRectangle(dpy, win, zap_gc, x + i * dx, lasty, dx, newh);
	i++;
	XFillRectangle(dpy, win, zap_gc, x + w - (i * dx), lasty, dx, newh);
	XSync(dpy, 0);
    }
    /* restore GC */
    XSetForeground(dpy, zap_gc, white);
}

/*
 * wipe across in a random direction
 */
static void
wipe(x, y, w, h)
    int         x,
                y,
                w,
                h;
{
    int         i,
                delta;

    delta = rnd(4) + 1;		/* 1-4 per swipe */
    if (debug)
	fprintf(stderr, "wiping %dx%d @ %d %d\n",
		w, h, x, y);
    switch (rnd(4)) {
    case 0:			/* top to bottom */
	for (i = 0; i < h; i += delta) {
	    XFillRectangle(dpy, win, zap_gc, x, y + i, w, delta);
	    XSync(dpy, 0);
	}
	break;
    case 1:			/* bottom to top */
	for (i = h - delta; i >= -delta; i -= delta) {
	    XFillRectangle(dpy, win, zap_gc, x, y + i, w, delta);
	    XSync(dpy, 0);
	}
	break;
    case 2:			/* left to right */
	for (i = 0; i < w; i += delta) {
	    XFillRectangle(dpy, win, zap_gc, x + i, y, delta, h);
	    XSync(dpy, 0);
	}
	break;
    case 3:			/* right to left */
	for (i = w - delta; i >= -delta; i -= delta) {
	    XFillRectangle(dpy, win, zap_gc, x + i, y, delta, h);
	    XSync(dpy, 0);
	}
	break;
    }
}

static void
zap_window(mode, x, y, w, h)
    int         mode;
    int         x,
                y,
                w,
                h;
{
    int         type;
    int         dx,
                dy;

    if (mode < 0)
	type = rnd(NUM_ZAP_TYPES);
    else
	type = mode;

    switch (type) {
    case ZAP_ZILLA:
	zilla(x, y, w, h);
	break;
    case ZAP_WIPE:
	wipe(x, y, w, h);
	break;
    case ZAP_TV_OFF:
	tv_off(x, y, w, h);
	break;
    case ZAP_FLING:
	/* delta is +-10 */
	dx = 10 - rnd(20);
	dy = 10 - rnd(20);
	fling_area(x, y, w, h, dx, dy);
	break;
    case ZAP_CLEAR_WIN:
    default:
	if (debug)
	    fprintf(stderr, "clearing %dx%d @ %d %d\n",
		    w, h, x, y);

#ifdef win_bg
	/*
	 * unless we snarf the screen BG (not likely) this should only be down
	 * for top level windows  - otherwise GC clip doesn't affect it
	 */
	XClearArea(dpy, win, x, y, w, h, False);
#else
	XFillRectangle(dpy, win, zap_gc, x, y, w, h);
#endif

	break;
    }
}

static void
zap_delay()
{
    XSync(dpy, 0);
    sleep(1);
}

static void
zap_all_windows(mode)
    int         mode;
{
    int         i;
    XRectangle  rect;
    Region      tmp,
                old,
                clip_region,
                full_screen;

    full_screen = rect_to_region(0, 0,
		      DisplayWidth(dpy, screen), DisplayHeight(dpy, screen));

    /* fill root area */
    clip_region = XCreateRegion();
    XXorRegion(full_screen, full_region, clip_region);
    XSetRegion(dpy, zap_gc, clip_region);
    XDestroyRegion(clip_region);
    XSetFillStyle(dpy, zap_gc, FillSolid);
    XSetForeground(dpy, zap_gc, white);
    XFillRectangle(dpy, win, zap_gc, 0, 0,
		   DisplayWidth(dpy, screen), DisplayHeight(dpy, screen));
    for (i = 0; i < num_windows; i++) {
	if (!region_list[i])
	    continue;

	if (check_escape_hatch()) {
	    printf("Victim found escape hatch -- sorry\n");
	    return;
	}
	/* figure out where zilla can go */
	clip_region = XCreateRegion();
	XXorRegion(full_screen, full_region, clip_region);
	XSetRegion(dpy, zilla_gc, clip_region);
	XDestroyRegion(clip_region);

	/* get bounds of window */
	XClipBox(region_list[i], &rect);

	/* compute region containing rest */
	tmp = region_list[i];
	region_list[i] = (Region) 0;

	/* if full didn't change, window is fully obscured */
	old = full_region;
	full_region = compute_full_region(region_list, num_windows);
	if (XEqualRegion(full_region, old)) {
	    XDestroyRegion(old);
	    XDestroyRegion(tmp);
	    continue;
	}
	XDestroyRegion(old);

	/* clip out rest */
	clip_region = XCreateRegion();
	XSubtractRegion(full_screen, full_region, clip_region);
	XSetRegion(dpy, zap_gc, clip_region);
	XDestroyRegion(clip_region);

	/* trash it */
	zap_window(mode, rect.x, rect.y, rect.width, rect.height);

	/* clean up */
	XDestroyRegion(tmp);

	/* wait */
	zap_delay();
    }
}
